
# This file was generated by 'versioneer.py' (0.18) from
# revision-control system data, or from the parent directory name of an
# unpacked source archive. Distribution tarballs contain a pre-generated copy
# of this file.

from __future__ import absolute_import
import json

version_json = '''
{
 "date": "2020-10-15T14:48:01-0700",
 "dirty": false,
 "error": null,
 "full-revisionid": "14c6a91c0696d02324452c6a02ce5db440d032b5",
 "version": "2.0.0"
}
'''  # END VERSION_JSON


def get_versions():
    return json.loads(version_json)
